/*-
 * Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ident "@(#) Copyright (C)2023..2024 @BABOLO http://www.babolo.ru/"
#ident "@(#) $Id: check.c,v 1.13 2024/06/15 16:22:29 babolo Exp $"

#define BLIN_COMPAT      4
#define Bpars_COMPAT     4
#define DIMENT_INTERNAL
#define _WITH_DPRINTF

#include <sys/types.h>
#include <sysexits.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include "../dimentia.h"

#define SHIFT_X (-4)
#define SHIFT_Y (-2)
#define PICSIZE_X 5
#define PICSIZE_Y 8
u_int64_t picmass[PICSIZE_Y][PICSIZE_X] =
{ {0x0201, 0x8001, 0xB001,      0,      0}
, {0xD000,      0,      0, 0x0070,      0}
, {     0, 0x0001,      0,      0,      0}
, {     0,      0,      0,      0,      0}
, {     0,      0, 0xA000, 0x0702, 0x2500}
, {     0,      0,      0,      0,      0}
, {     0,      0,      0,      0, 0x9000}
, {     0,      0,      0, 0x00F0,      0}
};

#define GLEVEL 2
#define UNIT_SIZE 24 /* <= 64 */
#define TILE_SIZE (1 << gsz[0])
const int gsz[] = {4, 4};

static void
/*****************************************************************************************************
                                                                                                  ****/
usage(int ex) {                                                                                  /****
                                                                                                  ****
 *****************************************************************************************************/
    fprintf( stderr
           , //"check @BABOLO V.M "VERS" http://www.babolo.ru/\n"
             "Usage: check [-options]\n"
             "   -B            border check\n"
             "   -R            read around\n"
             "   -Q file       stream pic\n"
             "   -v[v[v...]]   verbose\n"
             "   -q            silent\n"
             "   -= MODE       internal diagnostic\n"
             "   -?            this text\n"
           )
    ;
    exit(ex);
}

int main(int argc, char *argv[]) {
#   define isJX(EXCODE, FORMAT, ...) \
        do { ifBLIN_QX0("!!" FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
#   define isJW(EXCODE, FORMAT, ...) \
        do { ifBLIN_QW0("!!" FORMAT, ## __VA_ARGS__); ex = EXCODE; goto out; }while(0)
    const char *const opts = "BRQ:qv?=:";
    babolo_opts  *bos = NULL;              /*             */
    char          fl;                      /*                                       */
    const char   *optargs;                 /*                              */
    int           ex = EX_OK;              /*                                */
    int           fd = -1;                 /*   */
    diment_desc  *diment;
    u_int64_t     pixel = 0xBEAF;
    
    diment = diment_init(BLIN_GEN0 | DIMENT_TKNENT, UNIT_SIZE, 3, GLEVEL, gsz);
    if  (!diment) isJW(EX_SOFTWARE, "diment_init");
#   define blin_internal_flags (diment->flags & BLIN_MASK)
    diment_setfl(diment, &pixel);
    blin_ctl(BLIN_CTL_FLAG | BLIN_CTL_FSET, diment->flags);
#   define DO_BIG_READ     0x001000
#   define DO_BORDER_CHECK 0x002000
#   define DO_STEAM        0x004000
    bos = babolo_openopts(Bpars_NONU | Bpars_SYME | Bpars_ESYM, 'Z');
    if  (!bos) isJW(EX_SOFTWARE, "babolo_openopts");
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, opts)) isJW(EX_SOFTWARE, "babolo_setopts");
    while(!!(fl = babolo_getopts(bos))) {
        ifBLIN_QX2("flag %02X -%c", fl, fl);
        switch(fl) {
        case 'B':
            diment->flags |= DO_BORDER_CHECK;
            break;
        case 'R':
            diment->flags |= DO_BIG_READ;
            break;
        case 'Q':
            optargs = babolo_getoptsarg(bos);
            if  (!optargs) isJW(EX_USAGE, "no -Q arg");
            fd = open(optargs, O_WRONLY | O_APPEND | O_TRUNC | O_CREAT, 0666);
            if  (0 > fd) isJW(EX_OSERR, "open");
            diment->flags |= DO_STEAM;
            break;
        case 'q':
            BLIN_QUIET(diment->flags);
            break;
        case 'v':
            BLIN_VERBOSE(diment->flags);
            break;
        case '?':
            usage(EX_OK);
        case '=':
            switch((optargs = babolo_getoptsarg(bos))[0]) {
            case '0': blin_ctl(BLIN_CTL_DUMP);     exit(EX_OK);
            case '1': babolo_optest(opts);         exit(EX_OK);
            case '2': babolo_optext(bos);          exit(EX_OK);
            case '3': babolo_dumpopts(bos);        exit(EX_OK);
            case 'T': blin_ctl(BLIN_CTL_TCFN, ++optargs);    break;
            default : usage(EX_USAGE);
            }
            break;
        default:
            usage(EX_USAGE);
        }
        ifBLIN_QX2("-flag -%c", fl);
    }
    ifBLIN_QX2( "fill: x[%d, %d] y[%d, %d]"
              , diment->fill.xdown
              , diment->fill.xup
              , diment->fill.ydown
              , diment->fill.yup
              );
    ifBLIN_QX2("before all");
    pixel = 0x0EA0;
    ex = diment_un_cr(diment, 0, 0, &pixel);
    if  (!!ex) {
        ifBLIN_QX0("error c (0 0) %d\n", ex);
    }
    ex = diment_un_tk(diment, 0, 0, &pixel);
    if  (!!ex) {
        ifBLIN_QX0("error t (0 0) %d\n", ex);
    } else if(0x0EA0 != pixel) {
        ifBLIN_QX0("diff (0 0) %"BLIN_O"X\n", pixel);
    }
    
    ex = EX_OK;
    for (int y = -TILE_SIZE / 2; y < TILE_SIZE / 2; y++) {
        for (int x = -TILE_SIZE / 2; x < TILE_SIZE / 2; x++) {
            ex = diment_un_tk(diment, x, y, &pixel);
            if  (ex == ENOENT) {
                printf("VOID ");
            } else if(!!ex) {
                printf("ERR_ ");
            } else {
                printf("%4X ", (int)pixel);
            }
            ex = EX_OK;
        }
        printf("\n");
    }
    ifBLIN_QX2("previev done");
    for (int y = 0; y < PICSIZE_Y; y++) {
        for (int x = 0; x < PICSIZE_X; x++) {
            ex = diment_un_cr(diment, x + SHIFT_X, y + SHIFT_Y, &picmass[y][x]);
            if  (!!ex) isJX(EX_SOFTWARE, "diment_un_cr");
            ex = diment_un_tk(diment, x + SHIFT_X, y + SHIFT_Y, &pixel);
            if  (!!ex) isJX(EX_SOFTWARE, "diment_un_tk");
            
            if  (picmass[y][x] != pixel) {
                ifBLIN_QX0("wooloo %d %d", x, y);
            }
        }
    }
    ifBLIN_QX2( "writing done: x[%d, %d] y[%d, %d]"
              , diment->fill.xdown
              , diment->fill.xup
              , diment->fill.ydown
              , diment->fill.yup
              );
    for (int y = 0; y < PICSIZE_Y; y++) {
        for (int x = 0; x < PICSIZE_X; x++) {
            ex = diment_un_tk(diment, x + SHIFT_X, y + SHIFT_Y, &pixel);
            if  (!!ex) isJX(EX_SOFTWARE, "error (%d %d)\n", x, y);
            
            if  (picmass[y][x] != pixel) {
                printf("diff (%d %d) %4"BLIN_O"X %4"BLIN_O"X\n", x, y, pixel, picmass[y][x]);
            }
        }
    }
    ifBLIN_QX2("check read done");
    if  (!!(diment->flags & DO_BIG_READ)) {
        for (int y = -16; y < PICSIZE_Y + 16; y++) {
            for (int x = -16; x < PICSIZE_X + 16; x++) {
                ex = diment_un_tk(diment, x + SHIFT_X, y + SHIFT_Y, &pixel);
                if  (ex == ENOENT) {
                    printf("VOID ");
                } else if(!!ex) {
                    printf("ERR_ ");
                } else {
                    printf("%4X ", (int)pixel);
                }
                ex = EX_OK;
            }
            printf("\n");
        }
        ifBLIN_QX2("read done");
    }
    for (int y = 0; y < (1 << (gsz[GLEVEL - 1] * 2)); y += (1 << gsz[GLEVEL - 1])) {
        for (int x = 0; x < (1 << gsz[GLEVEL - 1]); x++) {
            printf(!diment->top.next[x + y].next ? "N " : "* ");
        }
        printf("\n");
    }
    ifBLIN_QX2("layer printed");
    if  (!!(diment->flags & DO_BORDER_CHECK)) {
        for (int iter = 0; iter < 20; iter++) {
            int num = (iter + 120);
            int tmp = (num + 128) % 256 - 128; /*     */
            
            pixel = 0x0F0F;
            ex = diment_un_cr(diment, num, num, &pixel);
            if  (!!ex) isJX(EX_SOFTWARE, "diment_un_cr %d", ex);
            for (int y = 0; y < (1 << (gsz[GLEVEL - 1] * 2)); y += (1 << gsz[GLEVEL - 1])) {
                for (int x = 0; x < (1 << gsz[GLEVEL - 1]); x++) {
                    printf(!diment->top.next[x + y].next ? "N " : "* ");
                }
                printf("\n");
            }
            ifBLIN_QX2("layer printed %d (%d) r(%d)", iter, num, tmp);
        }
        ifBLIN_QX2( "writing done: x[%d, %d] y[%d, %d]"
                  , diment->fill.xdown
                  , diment->fill.xup
                  , diment->fill.ydown
                  , diment->fill.yup
                  );
    }
    if  (!!(diment->flags & DO_STEAM)) {
        diment_doflag(diment, DIMENT_TKNENT, BLIN_CTL_FRST);
        if  (0 > dprintf( fd
                        , "P7\nWIDTH %d\nHEIGHT %d\nDEPTH 3\nMAXVAL 255\nTUPLTYPE RGB\nENDHDR\n"
                        , diment->fill.xup - diment->fill.xdown + 1
                        , diment->fill.yup - diment->fill.ydown + 1
                        )
            ) {
            isJW(EX_OSERR, "write");
        }
        ex = diment_flush(diment, diment->fill, fd, 3);
        if  (0 > ex) isJX(-ex, "diment_flush");
        ifBLIN_QX2("stream done");
        diment_doflag(diment, DIMENT_TKNENT, BLIN_CTL_FSET);
    }
out:
    babolo_closeopts(bos);
    diment_fini(diment);
    return(ex);
#   undef isJX
#   undef isJW
#   undef blin_internal_flags
}

/*
:
 2-3
 2-3
 
*/













